## ------------------------------------------------------------------------
##  $Id: Regexp.pm,v 1.1 2002/12/23 04:22:22 joko Exp $
## ------------------------------------------------------------------------
##  $Log: Regexp.pm,v $
##  Revision 1.1  2002/12/23 04:22:22  joko
##  + refactored from Data::Filter
##
##  Revision 1.1  2002/12/22 14:19:17  joko
##  + initial check-in
##
## ------------------------------------------------------------------------


package Data::Query::Filter::Regexp;

use strict;
use warnings;

use base 'DesignPattern::Object::Logger';


use Data::Dumper;

use Regexp::Group;
use libp qw( mkObject );
use Data::Compare::Struct qw( isEmpty );

# ------------    common perl object constructor    ------------
sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;
  my @args = ();
  @_ && (@args = @_);
  #$logger->debug( __PACKAGE__ . "->new( @args )" );      # this is not "common"!
  my $self = { @_ };
  bless $self, $class;
  $self->{caller} = caller;
  
  #print Dumper(caller(2));
  #exit;
  
  $self->_init();
  return $self;
}  

sub _init {
  my $self = shift;
  
  # try to load filter-declaration from configuration scope inside perl-module (yes - it's already abstracted out there!)
  $self->{declaration} = mkObject($self->{module});
#print Dumper($self->{declaration});
  

  # the regexp-object which does the hard work for us ;-)
  $self->{regexp} = Regexp::Group->new(
    'data' => \$self->{data},   # reference ...
    #'data' => $self->{data},   # ... or not?
    'metadata' => $self->{declaration}->metadata(),
    'patterns' => $self->{declaration}->patterns(),
    'coderefs' => $self->{declaration}->coderefs(),
    'verbose' => 1,
  );
  
}

sub run {
  my $self = shift;
  my $steps = shift;
  my $coderef = shift;

        #print "cb: $coderef", "\n";
  
  my $result = $self->{regexp}->scan($steps, $coderef);

  # configure tracing
    # respects additional trace-options passed to _trace-method
    $self->{TRACE_OPTIONS} = 1;
    # disables _any_ tracing - even if sub-conditions evaluate to true values
    $self->{TRACE_DISABLED} = 0;
  
  # trace the result
  # cute:  ;-)
  #$self->trace('matches after Regexp::Group->scan', $self->{regexp}->getMatches(), 1, undef, { tag => '', exit => 0 });
  #$self->trace('results after Regexp::Group->scan', $self->{regexp}->getResults(), 1, undef, { tag => '', exit => 0 });
  
  return $result;
}

sub index {
  my $self = shift;
  my $steps = shift;
  
  # configure tracing
    # respects additional trace-options passed to _trace-method
    $self->{TRACE_OPTIONS} = 1;
    # disables _any_ tracing - even if sub-conditions evaluate to true values
    $self->{TRACE_DISABLED} = 0;
  
  # trace the result
  # cute:  ;-)
  #$self->trace('matches after Regexp::Group->scan', $self->{regexp}->getMatches(), 1, undef, { tag => '', exit => 0 });
  #$self->trace('results after Regexp::Group->scan', $self->{regexp}->getResults(), 1, undef, { tag => '', exit => 0 });
  
  my $result = $self->{regexp}->scan($steps);


  return $result;
}

sub continue {
  my $self = shift;
  my $scankey = shift;
  my $result = $self->{regexp}->continue($scankey);
}

sub getResults {
  my $self = shift;
  return $self->{regexp}->{result};
}

sub getResultCount {
  my $self = shift;
  #my $count = ($#{$self->{regexp}->{result}} == -1 ? 0 : $#{$self->{regexp}->{result}});
  #$count++;

  
  my $matches = $self->{regexp}->getMatches();
  return if isEmpty($matches);
  
  my $count = $#{$matches};
  $count++;
  return $count;
}

1;
