##  ------------------------------------------------------------------------
##    $Id: NetLDAP.pm,v 1.2 2003/01/20 16:46:09 joko Exp $
##  ------------------------------------------------------------------------
##    $Log: NetLDAP.pm,v $
##    Revision 1.2  2003/01/20 16:46:09  joko
##    - doesn't use global '$ldap' as ldap-connection-handle any more - now using '$self->{_COREHANDLE}'
##    + sub getCOREHANDLE
##    + sub createChildNode
##
##  ------------------------------------------------------------------------


package Data::Storage::Handler::NetLDAP;

use strict;
use warnings;

#  Data::Storage::Handler
use base qw(
  Data::Storage::Handler::Abstract
  DesignPattern::Bridge
);


use Data::Dumper;
use Net::LDAP;
use Net::LDAP::Entry;
#use Net::LDAP::Search;

use Data::Identifier::Dn;

# get logger instance
my $logger = Log::Dispatch::Config->instance;


# the (package) global ldap-connection-handler
#my $self->{_COREHANDLE};

# TODO: implement this!
my $TRACELEVEL = 0;
my $logfile_errors = '../log/ldap_errors.log';


sub getMetaInfo {
  my $self = shift;
  $logger->debug( __PACKAGE__ . "->getMetaInfo()"  );
  return {
    'disconnectMethod' => 'disconnect',
  };
}

sub _init {
  my $self = shift;
#print "!!!!!!!!!!! _init", "\n";
  $self->load("Extensions");
}

sub connect {

  my $self = shift;

#print "connect!", "\n";

#print Dumper($self);

  my $dsn = $self->{locator}->{dsn};
  $logger->debug( __PACKAGE__ . "->connect( dsn $dsn )" );
  
  $self->{locator}->{connInfo} = $self->_getConnectionInfo($dsn);

  if (!$self->{locator}->{connInfo}->{host}) {
    $logger->critical( __PACKAGE__ . '->_getConnectionInfo: dsn-error: no host' );
    return;
  }
  $self->{locator}->{connInfo}->{port} ||= '389';
  
#print Dumper($self);

  $self->{_COREHANDLE} = Net::LDAP->new(
    $self->{locator}->{connInfo}->{host},
    port => $self->{locator}->{connInfo}->{port},
    #timeout => 120,
    debug => 0,
    #async => 1,
    #onerror => 'warn',
    onerror => \&_ldapError,
    #version => 3,
  ) 
  or do {
    my $message = {
      error => 'could not connect to server at ' . $self->{locator}->{connInfo}->{host} . ':' . $self->{locator}->{connInfo}->{port},
      code => 'none',
    };
    #$logger->error(__PACKAGE__ . "->connect: LDAP-ERROR: ");
    #$self->log("could not connect to ldap-server!");
    _ldapError($message);
    return;
  };
  
  #$self->{_COREHANDLE} = 
  $self->{_COREHANDLE}->bind( 
    $self->{locator}->{connInfo}->{binddn},
    password => $self->{locator}->{connInfo}->{pass}
  ) or die "$@";

  $self->{locator}->{status}->{connected} = 1;

#print "ok", "\n";

  return 1;

}

sub _getConnectionInfo {
  my $self = shift;
  my $dsn = shift;
  $dsn =~ m/host=(.+?);binddn=(.+?);pass=(.+?)$/;
  my $args = {
    host => $1,
    binddn => $2,
    pass => $3,
  };
  $args->{binddn} =~ s/'//g if $args->{binddn};
  return $args;
}

sub existsChildNode {
  my $self = shift;

#print "\n", "====  existsChildNode", "\n";
#exit;

  my $nodeName = shift;

  $logger->debug( __PACKAGE__ . "->existsChildNode( basedn='$self->{locator}->{basedn}', nodeName='$nodeName' )" );

  my $nodeDn = Data::Identifier::Dn->new(base => $self->{locator}->{basedn}, type => 'ou', name => $nodeName);
  my $nodeDn_asString = $nodeDn->asString();

  my $filter = "(objectClass=*)";

#{
  my $mesg = $self->{_COREHANDLE}->search(
    base => $nodeDn_asString,
    filter => $filter,
  );
#}

#return;

  #print "search-result-code: ", $mesg->code, "\n";
  #print "search-result-error: ", $mesg->error, "\n";
#  print "LDAP-error: ", $mesg->{errorMessage}, "\n" if $mesg->{errorMessage};
  
  #return;
  #print Dumper($mesg);
  #exit;
  
  #$mesg->code && die $mesg->error;

  return 1 if exists $mesg->{entries};
  return 0;

}

sub disconnect {
  my $self = shift;
  $self->{_COREHANDLE}->unbind;   # take down session
}

sub _ldapError {
  my $message = shift;
#print Dumper($message);
  my $textMessage = "LDAP-ERROR: " . $message->error . " - code: " . $message->code;
  $textMessage .= " - dn: " . $message->{matchedDN} if $message->{matchedDN};
  #print "msg: '$textMessage'", "\n";
  #a2f($logfile_errors, $textMessage) if !$TRACELEVEL;
  $logger->error($textMessage);
  return $message;
}

sub getCOREHANDLE {
  my $self = shift;
  return $self->{_COREHANDLE};
}

sub createChildNode {
  my $self = shift;
  my $nodeName = shift;
#print "createChildNode: $nodeName", "\n";
  my $nodeDn = Data::Identifier::Dn->new(base => $self->{locator}->{basedn}, type => 'ou', name => $nodeName);
  return $self->createDn($nodeDn);
}

1;
