##    $Id: API.pm,v 1.2 2003/01/20 16:59:48 joko Exp $
##
##    Copyright (c) 2002  Andreas Motl <andreas.motl@ilo.de>
##
##    See COPYRIGHT section in pod text below for usage and distribution rights.
##
##    ----------------------------------------------------------------------------------------
##    $Log: API.pm,v $
##    Revision 1.2  2003/01/20 16:59:48  joko
##    + cosmetics and debugging
##
##    Revision 1.1  2003/01/19 01:23:04  joko
##    + new from Data/Transfer/Sync.pm
##
##    ----------------------------------------------------------------------------------------


package Data::Transfer::Sync::API;

use strict;
use warnings;

use mixin::with qw( Data::Transfer::Sync );


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -   main

use Data::Dumper;

use Data::Compare::Struct qw( getDifference isEmpty );
use Data::Transform::Deep qw( merge );

# get logger instance
my $logger = Log::Dispatch::Config->instance;

sub _init {
  my $self = shift;
}

sub checkOptions {
  my $self = shift;
  my $opts = shift;
  
  my $result = 1;
  
  # check - do we have a target node?
  if (!$opts->{target_node}) {
    $logger->warning( __PACKAGE__ . "->checkOptions: Error while resolving resource metadata - no 'target node' could be determined.");
    $result = 0;
  }

  # check - do we have a mapping?
  if (!$opts->{mapping} && !$opts->{mapping_module}) {
    $logger->warning( __PACKAGE__ . "->checkOptions: Error while resolving resource metadata - no 'mapping' could be determined.");
    $result = 0;
  }
  
  return $result;
  
}

sub checkOptionsV2 {
  my $self = shift;

#print Dumper($self->{options});

  my $result = 1;

  # check - do we have a target node?
  if (!$self->{options}->{target}->{nodeName}) {
    $logger->warning( __PACKAGE__ . "->checkOptionsV2: No target given - please check metadata declaration.");
    $result = 0;
  }
  
  # check - do we have a mapping?
  if (!$self->{options}->{fieldmap}) {
    $logger->warning( __PACKAGE__ . "->checkOptionsV2: Error while resolving resource metadata - no 'fieldmap' could be determined.");
    $result = 0;
  }
  
  # TODO: extend!

  return $result;

}


sub prepareOptions {

  my $self = shift;
  my $opts = shift;

#print Dumper($opts);
#exit;

  $opts->{mode} ||= '';
  $opts->{erase} ||= 0;
  $opts->{prepare} ||= 0;
  #$opts->{import} ||= 0;
  
  if (!$opts->{source_node}) {
    $logger->error( __PACKAGE__ . "->prepareOptions failed: Please specify source-node!");
    return;
  }
  
  $logger->notice( __PACKAGE__ . "->prepareOptions( source_node $opts->{source_node} mode $opts->{mode} erase $opts->{erase} prepare $opts->{prepare} )");

  #if (!$opts->{mapping} || !$opts->{mapping_module}) {
  if (!$opts->{mapping}) {
    $logger->warning( __PACKAGE__ . "->prepareOptions: No mapping supplied - please check key 'mappings' in global configuration or specify additional argument '--mapping'.");
    #return;
  }

  $opts->{mapping_module} ||= $opts->{mapping};
  my $evstring = "use $opts->{mapping_module};";
  eval($evstring);
  if ($@) {
    $logger->warning( __PACKAGE__ . "->prepareOptions: error while trying to access mapping - $@");
    return;
  }

  # resolve mapping metadata (returned from sub)
  my $mapObject = $opts->{mapping_module}->new();
  #print Dumper($map);
  my $source_node_name = $opts->{source_node};
  # check if mapping for certain node is contained in mapping object
  if (!$mapObject->can($source_node_name)) {
    $logger->warning( __PACKAGE__ . "->prepareOptions: Can't access mapping for node \"$source_node_name\" - please check $opts->{mapping_module}.");
    return;
  }
  my $map = $mapObject->$source_node_name;

#print Dumper($map);

  # check here if "target" is actually a CODEref - in this case: resolve it - deprecated!!! ???
  if (ref $map->{target} eq 'CODE') {
    $map->{target} = $map->{target}->($source_node_name);
  }
  
  # resolve expressions (on nodename-level) here
  if ($map->{target} =~ m/^(code|expr):(.+?)$/) {
    my $target_dynamic_type = $1;
    my $target_dynamic_expression = $2;
    if (lc $target_dynamic_type eq 'code') {
#      $map->{target} = $mapObject->$target_dynamic_expression($map);
    }
  }

  # remove asymmetries from $map (patch keys)
  $map->{source_node} = $map->{source}; delete $map->{source};
  $map->{target_node} = $map->{target}; delete $map->{target};
  $map->{mapping} = $map->{details}; delete $map->{details};
  $map->{direction} = $map->{mode}; delete $map->{mode};

  # defaults (mostly for backward-compatibility)
  $map->{source_node} ||= $source_node_name;
  $map->{source_ident} ||= 'storage_method:id';
  $map->{target_ident} ||= 'property:oid';
  $map->{direction} ||= $opts->{mode};         # | PUSH | PULL | FULL
  $map->{method} ||= 'checksum';                # | timestamp
  $map->{source_exclude} ||= [qw( cs )];

  # merge map to opts
  map { $opts->{$_} = $map->{$_}; } keys %$map;
    
#print Dumper($opts);

  # TODO: move this to checkOptions...
  
  # check - do we have a target?
  if (!$opts->{target_node}) {
    $logger->warning( __PACKAGE__ . "->prepareOptions: No target given - please check metadata declaration.");
    return;
  }


  #return $opts;
  return 1;

}

sub prepareOptionsV2 {

  my $self = shift;
  my $opts = shift;

  # patch options
    $opts->{process}->{mode} ||= '';
    $opts->{process}->{prepare} ||= 0;
    $opts->{process}->{erase} ||= 0;

  # pre-check options
    if (!$self->_preCheckOptions($opts)) {
      $logger->error( __PACKAGE__ . "->prepareOptionsV2: _preCheckOptions failed.");
      return;
    }

  # inform user about option preparation
  $logger->notice( __PACKAGE__ . "->prepareOptionsV2( source.node='$opts->{source}->{nodeName}', target.node='$opts->{target}->{nodeName}', mode='$opts->{process}->{mode}', e='$opts->{process}->{erase}', p='$opts->{process}->{prepare}' )");

  # try to load mapping-metadata-container
  #my $mapObject = getNewPerlObjectByPkgName($opts->{map}->{moduleName});
  my $mapObject = DesignPattern::Object->fromPackage($opts->{map}->{moduleName});

  # try to resolve map from metadata-container

    # type of the item/node
    my $source_nodeType = $opts->{source}->{nodeType};
    
    # check if mapping for certain node is contained in mapping object
    if (!$mapObject->can($source_nodeType)) {
      $logger->warning( __PACKAGE__ . "->prepareOptionsV2: Can't access mapping for source-type=\"$source_nodeType\" - please check \"$opts->{map}->{moduleName}\".");
      return;
    }


    # get map
    my $map = $mapObject->$source_nodeType;
    #print Dumper($map);
  
=pod  
    # check here if "target" is actually a CODEref - in this case: resolve it - deprecated!!! ???
    if (ref $map->{target}->{address} eq 'CODE') {
      $map->{target}->{address} = $map->{target}->{address}->($source_nodeType);
    }
    
    # resolve expressions (on nodename-level) here
    if ($map->{target}->{address} =~ m/^(code|expr):(.+?)$/) {
      my $target_dynamic_type = $1;
      my $target_dynamic_expression = $2;
      if (lc $target_dynamic_type eq 'code') {
        $map->{target} = $mapObject->$target_dynamic_expression($map);
      }
    }
=cut

  #map { $opts->{$_} = $map->{$_}; } keys %$map;
  my $opts_merged = merge( $opts, $map );

  $self->{options} = $opts_merged;
  $self->{state}->{options_ready} = 1;

  return 1;

}


sub configure {
  my $self = shift;
  my @args = @_;
  if (!isEmpty(\@args)) {
    my %properties = @_;
    # merge args to properties
    map { $self->{$_} = $properties{$_}; } keys %properties;
    $self->_init();
    $self->_initV1();
  } else {
    #print "no args!", "\n";
  }
  #print Dumper($self);
  $self->{state}->{configured} = 1;
  return 1;
}

sub configureV2 {
  my $self = shift;
  my @args = @_;
  if (!isEmpty(\@args)) {
    my %properties = @_;
    # merge args to properties
    #map { $self->{$_} = $properties{$_}; } keys %properties;
    $self->{options} = merge($self->{options}, \%properties);
    $self->_init();
    #$self->_initV1();
  } else {
    #print "no args!", "\n";
  }

#print Dumper($self);

  $self->{state}->{configured} = 1;
  return 1;
}

sub _getDirectedArrow {
  my $self = shift;
  my $mode = shift;
  $mode ||= '';
  
  if (lc $mode eq 'push') {
    return '->';
  } elsif (lc $mode eq 'pull') {
    return '<-';
  } elsif (lc $mode eq 'full') {
    return '<->';
  } else {
    return '';
  }
}

# TODO: some feature to show off the progress of synchronization (cur/max * 100)
sub syncNodes {

  my $self = shift;
  my $args = shift;

  if (!$self->{state}->{configured}) {
    $logger->critical( __PACKAGE__ . "->syncNodes: Synchronization object is not configured/initialized correctly." );
    return;
  }

  # remember arguments through the whole processing
  $self->{args} = $args;

  $logger->debug( __PACKAGE__ . "->syncNodes: starting" );

  # hash to hold and/or fill in metadata required for the processing
  $self->{meta} = {};
  
  # hash to sum up results
  # TODO: re-implement! (sync-statistics???)

  # detect synchronization method to determine which optical symbol (directed arrow) to use
  my $mode = $self->{args}->{mode};   # V1
  $mode ||= $self->{options}->{process}->{mode};    # V2
  my $direction_arrow = $self->_getDirectedArrow($mode);

  if (!$self->{options}->{metadata}->{version} || $self->{options}->{metadata}->{version} < 0.2) {
    $self->_buildMetadataV1();
  } else {
    $self->_buildMetadataV2();
  }

  $logger->info( __PACKAGE__ . "->syncNodes: source=$self->{meta}->{source}->{dbKey}/$self->{meta}->{source}->{nodeName} [$self->{meta}->{source}->{nodeType}] $direction_arrow target=$self->{meta}->{target}->{dbKey}/$self->{meta}->{target}->{nodeName} [$self->{meta}->{target}->{nodeType}]" );

  return if !$self->_buildFieldmappingV1();
  return if !$self->_handleNodeContainers();
  return if !$self->_prepareSync();

#print Dumper($self);
#print Dumper($self->{args});
#print Dumper($self->{options});
#print Dumper($self->{meta});
#print Dumper($self->{metadata});

  $self->_syncNodes();
  
}


sub _handleNodeContainers {
  my $self = shift;

  # check partners/nodes: does partner exist / is node available?
  foreach my $partner (keys %{$self->{meta}}) {
    
    # 1. check partners & storages
    if (!$self->{meta}->{$partner}) {
      $logger->critical( __PACKAGE__ . "->_handleNodeContainers: Could not find partner '$partner' in configuration metadata." );
      next;
    }

    my $dbkey = $self->{meta}->{$partner}->{dbKey};

#print Dumper($self->{meta});

    if (!$self->{meta}->{$partner}->{storage}) {
      $logger->critical( __PACKAGE__ . "->_handleNodeContainers: Could not access storage ( partner='$partner', dbKey='$dbkey' ) - configuration-error?" );
      next;
    }
    
    # TODO:
    # 2. check if partners (and nodes?) are actually available....
    # eventually pre-check mode of access-attempt (read/write) here to provide an "early-croak" if possible
    
#    print Dumper($self->{meta}->{$partner}->{storage}->{locator});
  
    my $dbType = $self->{meta}->{$partner}->{storage}->{locator}->{type};
#print "dbType: $dbType", "\n";
    
    # 3. check nodes
    next if $dbType eq 'DBI';    # HACK for DBD::CSV - re-enable for others
    # get node-name
#      print Dumper($self);
    #print Dumper($self->{meta}->{$partner});
    #print "", $self->{meta}->{$partner}->{node}, "\n";
    my $nodename = $self->{meta}->{$partner}->{node};   # V1
    $nodename ||= $self->{meta}->{$partner}->{nodeName};    # V2
    # check if nodename is actually a CODEref, execute it to get a mapped/filtered target-nodename

#print "nodename: $nodename", "\n";

    $logger->debug( __PACKAGE__ . "->_handleNodeContainers: Accessing dbType=\"$dbType\", nodename=\"$nodename\"." );
    
=pod
    #print "----", ref $nodename, "\n";
    if ($nodename =~ m/CODE/) {
      print Dumper($self);
      #exit;
      $nodename = $nodename->($nodename);
    }
=cut    

#print Dumper($self);
    
    #print "partner: $partner - nodename: $nodename", "\n";
    
    if (!$self->{meta}->{$partner}->{storage}->existsChildNode($nodename)) {
#print "ex", "\n";
#exit;

      if ($partner eq 'target' && $self->{options}->{target}->{autocreateFolders}) {
        if (!$self->{meta}->{$partner}->{storage}->createChildNode($nodename)) {
          $logger->critical( __PACKAGE__ . "->_handleNodeContainers: Could not create node '$self->{meta}->{$partner}->{nodeName}\@$self->{meta}->{$partner}->{dbKey}' [$self->{meta}->{$partner}->{nodeType}]." );
          next;
        }
      } else {
        $logger->critical( __PACKAGE__ . "->_handleNodeContainers: Could not reach node \"$nodename\" at partner \"$partner\"." );
        next;
      }
    }
    
  }
  
  return 1;

}


sub _prepareSync {
  my $self = shift;

  # TODO:
  #   + if action == PUSH: start processing
  #  -+  if action == PULL: swap metadata and start processing
  #   -  if action == FULL: start processing, then swap metadata and (re-)start processing

  # manipulate metainfo according to direction of synchronization
  if (lc $self->{args}->{direction} eq 'push') {
    # just do it ...
  } elsif (lc $self->{args}->{direction} eq 'pull') {
    #print "=======SWAP", "\n";
    # swap
    ($self->{meta}->{source}, $self->{meta}->{target}) = 
        ($self->{meta}->{target}, $self->{meta}->{source});
  } elsif (lc $self->{args}->{direction} eq 'full') {
  } else {
  }

  # import flag means: prepare the source node to be syncable
  # this is useful if there are e.g. no "ident" or "checksum" columns yet inside a DBI like (row-based) storage
  if ($self->{args}->{prepare}) {
    $self->_prepareNode_MetaProperties('source');
    $self->_prepareNode_DummyIdent('source');
    #return;
    #$self->_erase_all($opts->{source_node});
  }
  
  # erase flag means: erase the target
  #if ($opts->{erase}) {
  if ($self->{args}->{erase}) {
    # TODO: move this method to the scope of the synchronization core and wrap it around different handlers
    #print "ERASE", "\n";
    $self->_erase_all('target');
  }
  
  return 1;

}

1;
