#################################
#
#  $Id: SyncStep.pm,v 1.1 2002/10/10 03:44:21 cvsjoko Exp $
#
#  $Log: SyncStep.pm,v $
#  Revision 1.1  2002/10/10 03:44:21  cvsjoko
#  + new
#
#
#################################

package Data::Transfer::SyncStep;

#require Exporter;
#@ISA = qw( Exporter );
#@EXPORT = qw( syncSet2Table );
                       
use strict;
use warnings;

use Data::Dumper;
use misc::HashExt;

# get logger instance
my $logger = Log::Dispatch::Config->instance;

sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;
  my $self = { @_ };
  $logger->debug( __PACKAGE__ . "->new(@_)" );
  return bless $self, $class;
}

sub _dumpCompact {
  $Data::Dumper::Indent = 0;
  my $result = Dumper(@_);
  $Data::Dumper::Indent = 2;
  return $result;
}

sub syncSet2Table {

  my $self = shift;
  my $args = shift;

  my $tc = OneLineDumpHash->new( {} );


  my @results;
  if ($args->{objectSet}) {
    # set of objects is already in $args
    $logger->debug( __PACKAGE__ . "->syncSet2Table( objectSet => '" . $args->{objectSet} . "' )" );
    @results = @{$args->{objectSet}};
  } else {
    # get set of objects from odbms by object name
    $logger->debug( __PACKAGE__ . "->syncSet2Table( objectName => '" . $args->{objectName} . "' )" );
    my $objects_set = $self->{L}->remote($args->{objectName});
    @results = $self->{L}->select($objects_set);
  }

  # iterate through set
  foreach my $obj (@results) {

    # clone object (in case we have to modify it here)
    my $objClone = $obj;
    
    # exclude defined fields  (simply delete from object)
    foreach my $exField ( @{$args->{exclude}} ) {
      delete $objClone->{$exField};
    }
    
    # calculate checksum for current object
    my $oid = $self->{L}->id($objClone);

    #my $objdump = $oid . "\n" . $objClone->quickdump();
    my $objdump = $oid . "\n" . Dumper($objClone);

    # TODO: $logger->dump( ... );
    #$logger->debug( __PACKAGE__ . ": " . $objdump );
    #$logger->dump( __PACKAGE__ . ": " . $objdump );
    
    # build checksum for current object-dump
      # md5-based fingerprint, base64 encoded (from Digest::MD5)
        #my $checksum_cur = md5_base64($objdump) . '==';
      # 32-bit integer "hash" value (maybe faster?) (from DBI)
        my $checksum_cur = DBI::hash($objdump, 1);

    # get checksum for current entry from rdbms
    my $sql = "SELECT cs FROM $args->{sqlTable} WHERE $args->{sqlIdField}='$oid';";
    #my $result = sendSql($sql);
    my $result = $self->{R}->sendCommand($sql);
    my $row = $result->getNextEntry();
    my $checksum_remote = $row->{cs};
    
    # determine if entry is "new" or "dirty"
    my $bool_new = !$checksum_remote;
    my $bool_dirty = $bool_new || $checksum_cur ne $checksum_remote || $args->{force};
    print "c" if $main::args->{verbose};
    if (!$bool_dirty) {
      $tc->{in_sync}++;
      next;
    }

    # field-structure for building sql
    # mapping of sql-fieldnames to object-attributes
      my $sqlmap;

      # manually set ...
        # ... object-id
        $sqlmap->{$args->{sqlIdField}} = $self->{L}->id($objClone);
        # ... checksum
        $sqlmap->{cs} = $checksum_cur;


      if ($args->{mapping}) {
        # apply mapping from $args->{mapping} to $sqlmap
        foreach my $key (keys %{$args->{mapping}}) {
          my $map_right = $args->{mapping}->{$key};
          #print "map: $map_right", "\n";
          my $value;
          # detect for callback
          if (ref($map_right) eq 'CODE') {
            $value = &$map_right($objClone);
          } else {
            $value = $objClone->{$map_right};
          }
          #$sqlmap->{$key} = $value;
          $sqlmap->{$key} = $self->{R}->quoteSql($value);
        }
      }

      if ($args->{mappingV2}) {
        
        # apply mapping from $args->{mappingV2} to $sqlmap
        foreach my $mapStep (@{$args->{mappingV2}}) {
          
          # prepare left/right keys/values
          my $left_key = $mapStep->{left};
          my $left_val = _resolveMapStepExpr( $objClone, $mapStep->{left} );
          my $right_key = $mapStep->{right};
          my $right_val = ( $mapStep->{right} );
          #print "map: $map_right", "\n";
          
          if ($mapStep->{method}) {
            if ($mapStep->{method} eq 'v:1') {
              $left_val = $left_key;
            }
          }

          #$sqlmap->{$key} = $value;
          #if ( grep(!/$right_key/, @{$args->{exclude}}) ) {
            $sqlmap->{$right_key} = $self->{R}->quoteSql($left_val);
          #}
        }
      }
      
    # TODO: $logger->dump( ... );
    #$logger->debug( "sqlmap:" . "\n" . Dumper($sqlmap) );

    my $sql_main;
    if ($bool_new) {
      $tc->{attempt_new}++;
      $sql_main = $self->{R}->hash2Sql($args->{sqlTable}, $sqlmap, 'SQL_INSERT');
    } else {
      $tc->{attempt_modify}++;
      $sql_main = $self->{R}->hash2Sql($args->{sqlTable}, $sqlmap, 'SQL_UPDATE', "$args->{sqlIdField}='$sqlmap->{$args->{sqlIdField}}'");
    }
    
    my $sqlHandle = $self->{R}->sendCommand($sql_main);
    if (!$sqlHandle->err) {
      $tc->{ok}++;
      print "t" if $main::args->{verbose};
    } else {
      $tc->{error}++;
      push( @{$tc->{error_per_row}}, { 
        statement => $sql_main,
        state => $sqlHandle->state, 
        err => $sqlHandle->err, 
        errstr => $sqlHandle->errstr, 
      } );
      #if ($args->{debug}) { print "sql-error with statement: $sql_main", "\n"; }
      print "e" if $main::args->{verbose};
    }
  
  }
  print "\n" if $main::args->{verbose};
  
  
  # build user-message from some stats and log it via "sysevent"
  # change logging level in case of an error
  
    my $level = 'LEVEL_INFO';
    
    my $msg;
    
#    $msg .= "stats: ";
#      $tc->{in_sync} && ($msg .= "in-sync: $tc->{in_sync}");
#      $tc->{attempt_new} && ($msg .= "  /  " . "attempt-new: $tc->{attempt_new}");
#      $tc->{attempt_modify} && ($msg .= "  /  " . "attempt-modify: $tc->{attempt_modify}");
#      $tc->{ok} && ($msg .= "  /  " . "result-ok: $tc->{ok}");
#      $tc->{error} && ($msg .= "  /  " . "result-error: $tc->{error}");
    
    $msg .= "stats: $tc";
    
    if ($tc->{error_per_row}) {
      $msg .= "\n";
      $level = 'LEVEL_NOTIFY';
      $msg .= "errors:" . "\n";
      $msg .= Dumper($tc->{error_per_row});
    }
    
    # todo!!!
    #sysevent( { usermsg => $msg, level => $level }, $taskEvent );
    $logger->info($msg);

  return $tc;

}

sub _resolveMapStepExpr {

  my $obj = shift;
  my $expr = shift;

          my $value;
          # detect for callback (code-reference)
          if (ref($expr) eq 'CODE') {
            $value = &$expr($obj);
          } elsif ($expr =~ m/->/) {
            # use expr as complex object reference declaration (obj->subObj->subSubObj->0->attribute)
            my $objPerlRefString = _mapStepExprToPerlRefString($expr);
            #print "\n", "expr: $expr";
            #print "\n", "objPerlRefString: $objPerlRefString";
            $value = eval('$obj' . '->' . $objPerlRefString);
          } else {
            # use expr as simple scalar key (attributename)
            $value = $obj->{$expr};
          }
          
  return $value;
  
}

sub _mapStepExprToPerlRefString {
  my $expr = shift;

  # split expression by dereference operators first
  my @parts = split(/->/, $expr);

  # wrap []'s around each part, if it consists of numeric characters only (=> numeric array-index),
  # use {}'s, if there are word-characters in it (=> alphanumeric hash-key)
  map {
    m/^\d+$/ && ($_ = "[$_]") || ($_ = "{$_}");
  } @parts;

  # join parts with dereference operators together again and return built string
  return join('->', @parts);
}


1;